'use strict';

function comparatorFactoryFactory(rule) {
    return function () {
        return addComparatorFeatures(createComparator(rule || {}, arguments));
    };
}
// avoid trivial troubles on import
comparatorFactoryFactory.default = comparatorFactoryFactory;
function addComparatorFeatures(comparator) {
    comparator.reversed = function (really) {
        if (really === void 0) { really = true; }
        return really ? addComparatorFeatures(function (a, b) { var result = comparator(a, b); return result && -result; }) : comparator;
    };
    comparator.or = function (anotherComparator) { return addComparatorFeatures(function (a, b) { return comparator(a, b) || anotherComparator(a, b); }); };
    return comparator;
}
function createComparator(rule, keys) {
    var comparators = [];
    for (var i = 0; i < keys.length; ++i) {
        comparators.push(_createComparator(rule, keys[i]));
    }
    switch (comparators.length) {
        case 0: return _createComparator(rule);
        case 1: return comparators[0];
        case 2: return function (a, b) { return comparators[0](a, b) || comparators[1](a, b); };
        case 3: return function (a, b) { return comparators[0](a, b) || comparators[1](a, b) || comparators[2](a, b); };
        default: return function (a, b) {
            var result = 0;
            for (var i = 0; i < comparators.length && (result = comparators[i](a, b)) === 0; ++i)
                ;
            return result;
        };
    }
}
function _createComparator(rule, key) {
    var valueSelector = key ? (rule.selector || defaultSelector)(key) : function (obj) { return obj; };
    var compareSpecials = createSpecialsComparator(rule.specials);
    var collator = rule.collator;
    var compareString = collator && typeof collator.compare === 'function' ? collator : Intl.Collator(collator && collator.locales, collator);
    var compareValue = function (a, b) {
        var result;
        return (a === b ? 0 :
            (result = compareSpecials && compareSpecials(a, b)) !== undefined ? result :
                a === undefined ? -1 :
                    b === undefined ? 1 :
                        a === null ? -1 :
                            b === null ? 1 :
                                a !== a ? b !== b ? 0 : -1 : // NaN !== NaN
                                    b !== b ? 1 :
                                        Array.isArray(a) && Array.isArray(b) ? compareArray(a, b, compareValue) :
                                            typeof a === 'string' || typeof b === 'string' ? compareString.compare(String(a), String(b)) :
                                                a < b ? -1 : 1);
    };
    return function (a, b) { return compareValue(valueSelector(a), valueSelector(b)); };
}
function createSpecialsComparator(specials) {
    if (!specials || !specials.length) {
        return;
    }
    var comparators = specials.map(function (_a) {
        var special = _a[0], handling = _a[1];
        var isSpecial = typeof special === 'function' ? special :
            special !== special ? function (value) { return value !== value; } :
                function (value) { return value === special; };
        var leftResult = handling === 'last' ? 1 : -1;
        return function (a, b) { return isSpecial(a) ? isSpecial(b) ? 0 : leftResult : isSpecial(b) ? -leftResult : undefined; };
    });
    return function (a, b) {
        var result;
        for (var i = 0; i < comparators.length && (result = comparators[i](a, b)) === undefined; ++i)
            ;
        return result;
    };
}
function compareArray(a, b, compare) {
    var result = 0;
    var length = Math.min(a.length, b.length);
    for (var i = 0; i < length && !(result = compare(a[i], b[i])); ++i)
        ;
    return result || compare(a.length, b.length);
}
function defaultSelector(key) {
    if (typeof key !== 'function') {
        throw new Error('comparison key must be a function');
    }
    else {
        return function (obj) { try {
            return key(obj);
        }
        catch (_a) { } };
    }
}

module.exports = comparatorFactoryFactory;
